from __future__ import annotations

from nefertem_core.metadata.report import NefertemBaseReport


class NefertemProfile(NefertemBaseReport):
    """
    Succint version of a profile produced by some profiling library.

    Attributes
    ----------
    stats : dict
        Descriptors of data stats.
    fields : dict
        Descriptors of data fields.
    """

    def __init__(
        self,
        framework_name: str,
        framework_version: str,
        duration: float,
        stats: dict,
        fields: dict,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(framework_name, framework_version, duration)
        self.stats = stats
        self.fields = fields



"""
Profiling plugin abstract class module.
"""
from __future__ import annotations

from abc import abstractmethod

from nefertem_core.plugins.builder import PluginBuilder
from nefertem_metric.plugins.metric import Metric


class MetricPluginBuilder(PluginBuilder):
    """
    Profiling plugin builder.
    """

    @staticmethod
    @abstractmethod
    def _filter_metrics(metrics: list[Metric]) -> list[Metric]:
        """
        Filter metric by library.
        """



from nefertem_core.metadata.report import NefertemBaseReport


class ProfileMetric:
    """
    Metric report data.

    Attributes
    ----------
    name: str
        Metric identificator as defined by the corresponding domain.
    title: str
        Human readable name for the metric.
    type: str
        Metric domain (es., library).
    args: dict
        Metric arguments (if applies)
    value: Any
        Metric value

    """

    def __init__(self, name: str, title: str, type: str, args: dict, value: any) -> None:
        """
        Constructor.
        """
        self.name = name
        self.title = title
        self.type = type
        self.args = args
        self.value = value

    def to_dict(self) -> dict:
        return self.__dict__


class NefertemMetricReport(NefertemBaseReport):
    """
    Succint version of a profile produced by some profiling library.

    Attributes
    ----------
    stats : dict
        Descriptors of data stats.
    fields : dict
        Descriptors of data fields.
    metrics : list[dict]
        List of metric evaluations over dataset.
    field_metrics: dict
        Dataset field metrics
    """

    def __init__(
        self,
        framework_name: str,
        framework_version: str,
        duration: float,
        stats: dict,
        fields: dict,
        metrics: list[dict],
        field_metrics: dict,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(framework_name, framework_version, duration)
        self.stats = stats
        self.fields = fields
        self.metrics = metrics
        self.field_metrics = field_metrics



"""
Validation plugin utils.
"""
from __future__ import annotations


def render_error_type(code: str) -> dict:
    """
    Return standard errors record format.

    Parameters
    ----------
    code : str
        Error code.

    Returns
    -------
    dict
        Error type record.
    """
    return {"type": code}


def parse_error_report(error_list: list, report_type) -> list:
    """
    Return a list of record according to user parameter.

    Parameters
    ----------
    error_list : list
        List of errors.

    Returns
    -------
    list
        List of errors.
    """
    if report_type == "count":
        return []
    if report_type == "partial":
        if len(error_list) <= 100:
            return error_list
        return error_list[:100]
    if report_type == "full":
        return error_list


def get_errors(count: int = 0, records: list = None) -> dict:
    """
    Return a common error structure.

    Parameters
    ----------
    count : int
        Number of errors.
    records : list
        List of errors.

    Returns
    -------
    dict
        Error structure.
    """
    if records is None:
        records = []
    return {"count": count, "records": records}



"""
Nefertem Inference Plugin Utils.
"""
from __future__ import annotations


def get_fields(name: str = None, type_: str = None) -> dict:
    """
    Return a common field structure.

    Parameters
    ----------
    name : str
        Field name.
    type_ : str
        Field type.

    Returns
    -------
    dict
        Field structure.
    """
    if name is None:
        name = "unknown"
    if type_ is None:
        type_ = "unknown"
    return {"name": name, "type": type_}



from __future__ import annotations

from nefertem_core.utils.utils import build_uuid
from pydantic import BaseModel, Field


class Constraint(BaseModel):
    """
    Base model for constraint.
    """

    id: str = Field(default_factory=build_uuid)
    """UUID of constraint."""

    name: str
    """Constraint id."""

    type: str
    """Constraint type."""

    title: str
    """Human readable name for the constraint."""

    resources: list[str]
    """List of resources affected by the constraint."""

    weight: int
    """Criticity of an eventual error encountered in the validation for the constraint."""



from __future__ import annotations

from nefertem_core.metadata.report import NefertemBaseReport


class NefertemSchema(NefertemBaseReport):
    """
    Succint version of an inferred schema produced by some inference library.

    Attributes
    ----------
    fields : list
        A list of fields.

    """

    def __init__(
        self,
        framework_name: str,
        framework_version: str,
        duration: float,
        fields: list,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(framework_name, framework_version, duration)
        self.fields = fields



"""
Run handler module.
"""
from __future__ import annotations

from typing import Any

from nefertem_core.metadata.blob import Blob
from nefertem_core.plugins.utils import ResultType
from nefertem_core.run.run import Run
from nefertem_validation.metadata.report import NefertemReport


class RunValidation(Run):
    """
    Run validation extension.

    Methods
    -------
    validate
        Execute validation on resources.
    validate_framework
        Execute validation on resources with validation frameworks.
    validate_nefertem
        Execute validation on resources with Nefertem.
    log_report
        Log NefertemReport.
    persist_report
        Persist frameworks reports.
    """

    def validate_framework(self, constraints: list[dict], error_report: str | None = "partial") -> list[Any]:
        """
        Execute validation on resources with validation frameworks.

        Parameters
        ----------
        constraints : list[dict]
            list of constraint to validate resources.
        error_report : str
            Flag to render the error output of the nefertem report.
            Accepts 'count', 'partial' or 'full'.
            'count' returns the errors count of the validation,
            'partial' return the errors count and a list of errors (max 100),
            'full' returns errors count and the list of all encountered errors.

        Returns
        -------
        list[Any]
            Return a list of framework results.
        """
        reports = self.run_handler.get_item(ResultType.FRAMEWORK.value)
        if reports:
            return reports

        self.run_handler.run(self.run_info.resources, constraints, error_report)
        return self.run_handler.get_item(ResultType.FRAMEWORK.value)

    def validate_nefertem(self, constraints: list[dict], error_report: str | None = "partial") -> list[NefertemReport]:
        """
        Execute validation on resources with Nefertem.

        Parameters
        ----------
        constraints : list[dict]
            list of constraint to validate resources.
        error_report : str
            Flag to render the error output of the nefertem report.
            Accepts 'count', 'partial' or 'full'.
            'count' returns the errors count of the validation,
            'partial' return the errors count and a list of errors (max 100),
            'full' returns errors count and the list of all encountered errors.

        Returns
        -------
        list[NefertemReport]
            Return a list of "NefertemReport".
        """
        reports = self.run_handler.get_item(ResultType.NEFERTEM.value)
        if reports:
            return reports

        self.run_handler.run(self.run_info.resources, constraints, error_report)
        return self.run_handler.get_item(ResultType.NEFERTEM.value)

    def validate(self, constraints: list[dict], error_report: str | None = "partial") -> Any:
        """
        Execute validation on resources.

        Parameters
        ----------
        constraints : list[dict]
            list of constraint to validate resources.
        error_report : str
            Flag to render the error output of the nefertem report.
            Accepts 'count', 'partial' or 'full'.
            'count' returns the errors count of the validation,
            'partial' return the errors count and a list of errors (max 100),
            'full' returns errors count and the list of all encountered errors.
        only_nt : bool
            Flag to return only the Nefertem report.

        Returns
        -------
        Any
            Return a list of "NefertemReport" and the corresponding list of framework results.

        """
        return self.validate_framework(constraints, error_report), self.validate_nefertem(constraints, error_report)

    def log_report(self) -> None:
        """
        Log NefertemReport.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.NEFERTEM.value):
            metadata = Blob(*self._get_base_args(), obj.object.to_dict()).to_dict()
            self._log_metadata(metadata, obj.filename)

    def persist_report(self) -> None:
        """
        Persist frameworks reports.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.RENDERED.value):
            self._persist_artifact(obj.object, obj.filename)



"""
Run handler module.
"""
from __future__ import annotations

from typing import Any

from nefertem_core.metadata.blob import Blob
from nefertem_core.plugins.utils import ResultType
from nefertem_core.run.run import Run
from nefertem_metric.metadata.report import NefertemMetricReport


class RunMetric(Run):
    """
    Run metric extension.

    Methods
    -------
    metric
        Execute metric on resources.
    metric_framework
        Execute metric on resources with metric frameworks.
    metric_nefertem
        Execute metric on resources with Nefertem.
    log_metric
        Log NefertemMetricReports.
    persist_metric
        Persist frameworks metrics.
    """

    ############################
    # Metric
    ############################

    def metric_framework(self, metrics: list[dict]) -> list[Any]:
        """
        Execute metric on resources with metric frameworks.

        Parameters
        ----------
        metrics: list[dict]
            Optional list of metrics to evaluate over resources.

        Returns
        -------
        list[Any]
            Return a list of framework results.

        """
        metrics = self.run_handler.get_item(ResultType.FRAMEWORK.value)
        if metrics:
            return metrics

        self.run_handler.run(self.run_info.resources, metrics)
        return self.run_handler.get_item(ResultType.FRAMEWORK.value)

    def metric_nefertem(self, metrics: list[dict]) -> list[NefertemMetricReport]:
        """
        Execute metric on resources with Nefertem.

        Parameters
        ----------
        metrics: list[dict]
            Optional list of metrics to evaluate over resources.

        Returns
        -------
        list[NefertemMetricReport]
            Return a list of NefertemMetricReport.

        """
        metrics = self.run_handler.get_item(ResultType.NEFERTEM.value)
        if metrics:
            return metrics

        self.run_handler.run(self.run_info.resources, metrics)
        return self.run_handler.get_item(ResultType.NEFERTEM.value)

    def metric(self, metrics: list[dict]) -> tuple[list[Any], list[NefertemMetricReport]]:
        """
        Execute metric on resources.

        Parameters
        ----------
        metrics: list[dict]
            Optional list of metrics to evaluate over resources.

        Returns
        -------
        Any
            Return a list of NefertemMetricReport and the corresponding list of framework results.

        """
        return self.metric_framework(metrics), self.metric_nefertem(metrics)

    def log_metric(self) -> None:
        """
        Log NefertemMetricReports.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.NEFERTEM.value):
            metadata = Blob(*self._get_base_args(), obj.object.to_dict()).to_dict()
            self._log_metadata(metadata, obj.filename)

    def persist_metric(self) -> None:
        """
        Persist frameworks metrics.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.RENDERED.value):
            self._persist_artifact(obj.object, obj.filename)



"""
Run handler module.
"""
from __future__ import annotations

from typing import Any

from nefertem_core.metadata.blob import Blob
from nefertem_core.plugins.utils import ResultType
from nefertem_core.run.run import Run
from nefertem_profiling.metadata.report import NefertemProfile


class RunProfiling(Run):
    """
    Run profiling extension.

    Methods
    -------
    profile
        Execute profile profiling on resources.
    profile_framework
        Execute profile profiling on resources with profiling frameworks.
    profile_nefertem
        Execute profile profiling on resources with Nefertem.
    log_profile
        Log NefertemProfiles.
    persist_profile
        Persist frameworks profiles.
    """

    def profile_framework(self) -> list[Any]:
        """
        Execute profiling on resources with profiling frameworks.

        Returns
        -------
        list[Any]
            Return a list of framework results.

        """
        profiles = self.run_handler.get_item(ResultType.FRAMEWORK.value)
        if profiles:
            return profiles

        self.run_handler.run(self.run_info.resources)
        return self.run_handler.get_item(ResultType.FRAMEWORK.value)

    def profile_nefertem(self) -> list[NefertemProfile]:
        """
        Execute profiling on resources with Nefertem.

        Returns
        -------
        list[NefertemProfile]
            Return a list of NefertemProfile.

        """
        profiles = self.run_handler.get_item(ResultType.NEFERTEM.value)
        if profiles:
            return profiles

        self.run_handler.run(self.run_info.resources)
        return self.run_handler.get_item(ResultType.NEFERTEM.value)

    def profile(self) -> tuple[list[Any], list[NefertemProfile]]:
        """
        Execute profiling on resources.

        Returns
        -------
        Any
            Return a list of NefertemProfile and the corresponding list of framework results.

        """
        return self.profile_framework(), self.profile_nefertem()

    def log_profile(self) -> None:
        """
        Log NefertemProfiles.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.NEFERTEM.value):
            metadata = Blob(*self._get_base_args(), obj.object.to_dict()).to_dict()
            self._log_metadata(metadata, obj.filename)

    def persist_profile(self) -> None:
        """
        Persist frameworks profiles.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.RENDERED.value):
            self._persist_artifact(obj.object, obj.filename)



"""
Profiling plugin abstract class module.
"""
from __future__ import annotations

from nefertem_core.plugins.builder import PluginBuilder


class ProfilingPluginBuilder(PluginBuilder):
    """
    Profiling plugin builder.
    """



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from nefertem_core.plugins.plugin import Plugin
from nefertem_core.plugins.utils import ResultType


class ProfilingPlugin(Plugin):
    """
    Run plugin that executes profiling over a Resource.
    """

    def execute(self) -> dict:
        """
        Method that call specific execution.
        """
        plugin = f"Plugin: {self.framework_name()} {self.id};"
        self.logger.info(f"Execute profiling - {plugin}")
        lib_result = self.profile()
        self.logger.info(f"Render report - {plugin}")
        nt_result = self.render_nefertem(lib_result)
        self.logger.info(f"Render artifact - {plugin}")
        render_result = self.render_artifact(lib_result)
        return {
            ResultType.FRAMEWORK.value: lib_result,
            ResultType.NEFERTEM.value: nt_result,
            ResultType.RENDERED.value: render_result,
            ResultType.LIBRARY.value: self.get_framework(),
        }

    @abstractmethod
    def profile(self) -> Any:
        """
        Generate a data profile.
        """



from __future__ import annotations

from nefertem_core.metadata.report import NefertemBaseReport


class NefertemReport(NefertemBaseReport):
    """
    Succint version of a report produced by some validation library.

    Attributes
    ----------
    constraint : dict
        Constraint validated.
    valid : bool
        Validation outcome.
    errors : list
        List of errors found by validation process.

    """

    def __init__(
        self,
        framework_name: str,
        framework_version: str,
        duration: float,
        constraint: dict,
        valid: bool,
        errors: list,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(framework_name, framework_version, duration)
        self.constraint = constraint
        self.valid = valid
        self.errors = errors



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from nefertem_core.plugins.plugin import Plugin
from nefertem_core.plugins.utils import ResultType


class MetricPlugin(Plugin):
    """
    Run plugin that executes metric over a Resource.
    """

    _fn_metric = "metric_{}"

    def __init__(self) -> None:
        """
        Constructor.
        """
        super().__init__()
        self.metric = None

    def execute(self) -> dict:
        """
        Method that call specific execution.
        """
        plugin = f"Plugin: {self.framework_name()} {self.id};"
        self.logger.info(
            f"Execute metric - {plugin}" + (" Metric: {self.metric.name}" if self.metric is not None else "")
        )
        lib_result = self.measure()
        self.logger.info(f"Render report - {plugin}")
        nt_result = self.render_nefertem(lib_result)
        self.logger.info(f"Render artifact - {plugin}")
        render_result = self.render_artifact(lib_result)
        return {
            ResultType.FRAMEWORK.value: lib_result,
            ResultType.NEFERTEM.value: nt_result,
            ResultType.RENDERED.value: render_result,
            ResultType.LIBRARY.value: self.get_framework(),
        }

    @abstractmethod
    def measure(self) -> Any:
        """
        Generate metrics.
        """



from __future__ import annotations

import typing
from abc import abstractmethod

from nefertem_core.plugins.builder import PluginBuilder

if typing.TYPE_CHECKING:
    from nefertem_validation.plugins.constraint import Constraint


class ValidationPluginBuilder(PluginBuilder):
    """
    Validation plugin builder.
    """

    @staticmethod
    @abstractmethod
    def _filter_constraints(constraints: list[dict]) -> list[Constraint]:
        """
        Filter constraints by library.
        """



"""
Run handler module.
"""
from __future__ import annotations

from typing import Any

from nefertem_core.metadata.blob import Blob
from nefertem_core.plugins.utils import ResultType
from nefertem_core.run.run import Run
from nefertem_inference.metadata.report import NefertemSchema


class RunInference(Run):
    """
    Run inference extension.

    Methods
    -------
    infer
        Execute schema inference on resources.
    infer_framework
        Execute schema inference on resources with inference frameworks.
    infer_nefertem
        Execute schema inference on resources with Nefertem.
    log_schema
        Log NefertemSchemas.
    persist_schema
        Persist frameworks schemas.
    persist_data
        Persist input data as artifacts into default store.
    """

    def infer_framework(self) -> list[Any]:
        """
        Execute schema inference on resources with inference frameworks.

        Returns
        -------
        list[Any]
            Return a list of framework results.

        """
        schemas = self.run_handler.get_item(ResultType.FRAMEWORK.value)
        if schemas:
            return schemas

        self.run_handler.run(self.run_info.resources)
        return self.run_handler.get_item(ResultType.FRAMEWORK.value)

    def infer_nefertem(self) -> list[NefertemSchema]:
        """
        Execute schema inference on resources with Nefertem.

        Returns
        -------
        list[NefertemSchema]
            Return a list of NefertemSchemas.

        """
        schemas = self.run_handler.get_item(ResultType.NEFERTEM.value)
        if schemas:
            return schemas

        self.run_handler.run(self.run_info.resources)
        return self.run_handler.get_item(ResultType.NEFERTEM.value)

    def infer(self) -> tuple[list[Any], list[NefertemSchema]]:
        """
        Execute schema inference on resources.

        Returns
        -------
        Any
            Return a list of NefertemSchemas and the corresponding list of framework results.
        """
        return self.infer_framework(), self.infer_nefertem()

    def log_schema(self) -> None:
        """
        Log NefertemSchemas.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.NEFERTEM.value):
            metadata = Blob(*self._get_base_args(), obj.object.to_dict()).to_dict()
            self._log_metadata(metadata, obj.filename)

    def persist_schema(self) -> None:
        """
        Persist frameworks schemas.

        Returns
        -------
        None
        """
        for obj in self.run_handler.get_item(ResultType.RENDERED.value):
            self._persist_artifact(obj.object, obj.filename)



"""
Validation plugin abstract class module.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any

from nefertem_core.plugins.plugin import Plugin
from nefertem_core.plugins.utils import ResultType


class ValidationPlugin(Plugin):
    """
    Run plugin that executes validation over a Resource.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        super().__init__()
        self.constraint = None
        self.error_report = None

    def execute(self) -> dict:
        """
        Method that call specific execution.

        Returns
        -------
        dict
            Results of execution.
        """
        plugin = f"Plugin: {self.framework_name()} {self.id};"
        constraint = f"Constraint: {self.constraint.name};"
        resources = f"Resources: {self.constraint.resources};"
        self.logger.info(f"Execute validation - {plugin} {constraint} {resources}")
        lib_result = self.validate()
        self.logger.info(f"Render report - {plugin}")
        nt_result = self.render_nefertem(lib_result)
        self.logger.info(f"Render artifact - {plugin}")
        render_result = self.render_artifact(lib_result)
        return {
            ResultType.FRAMEWORK.value: lib_result,
            ResultType.NEFERTEM.value: nt_result,
            ResultType.RENDERED.value: render_result,
            ResultType.LIBRARY.value: self.get_framework(),
        }

    @abstractmethod
    def validate(self) -> Any:
        """
        Validate a resource.
        """



from __future__ import annotations

from nefertem_core.utils.utils import build_uuid
from pydantic import BaseModel, Field


class Metric(BaseModel):
    """
    Base model for metric to be evaluated via profiling.
    """

    id: str = Field(default_factory=build_uuid)
    """UUID of metric."""

    name: str
    """Metric identificator as defined by the corresponding domain."""

    title: str
    """Human readable name for the metric."""

    resources: list[str]
    """List of resources on which the metric should be evaluated."""
